#include "definitions.h"
#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>


#include <util/delay.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "uart.h"
#include "uexpect.h"

#define COUNTER_RED_LED OCR2B //Output Compare Register 2A (8 Bit)
#define COUNTER_BLUE_LED OCR0A //Output Compare Register 0A (16 Bit, PWM aber 8 Bit)
#define COUNTER_GREEN_LED OCR0B //Output Compare Register 0B (16 Bit, PWM aber 8 Bit)

#define ARRAY_SIZE(array) (sizeof(array) / sizeof(*array))

FILE uart_output = FDEV_SETUP_STREAM(uart_putchar, NULL, _FDEV_SETUP_WRITE); 
FILE uart_input = FDEV_SETUP_STREAM(NULL, uart_getchar, _FDEV_SETUP_READ);
FILE uart_io = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);


typedef struct{
	//Array of chars
	uint8_t buffer[BUF_SIZE];
	//array element index
	uint8_t index;
	//pointer to head 
	uint8_t head; 
}u8buf;

//declare buffer
u8buf buf;


char* responseOk = "\r\nOK\r\n";

void setRGB(int r, int g, int b) {
	COUNTER_BLUE_LED = b;
	COUNTER_GREEN_LED = g;
	COUNTER_RED_LED = r;
}

void setR() {
	setRGB(255,0,0);
}

void setG() {
	setRGB(0, 255, 0);
}

void setB() {
	setRGB(0, 0, 255);
}

//initialize buffer
void BufferInit(u8buf *buf) {
	//set index to start of buffer
	buf->index=0;
	buf->head=0;
	buf->buffer[0] = '\0';
}


//write to buffer routine
uint8_t BufferWrite(u8buf *buf, uint8_t u8data) {
	if (buf->index < BUF_SIZE)
	{
		buf->buffer[buf->index] = u8data;
		//increment buffer index
		buf->index++;
		return 0;
	}
	else return 1;
}

void blinkR(int times) {
	for(int i=0; i < times; i++) {
		setR();
		_delay_ms(200);
		setRGB(0,0,0);
		_delay_ms(200);
	}
}

void blinkG(int times) {
	for(int i=0; i < times; i++) {
		setG();
		_delay_ms(100);
		setRGB(0,0,0);
		_delay_ms(100);
	}
}

uint8_t BufferRead(u8buf *buf, volatile uint8_t *u8data) {
	if(buf->index > 0)
	{
		buf->index--;
		*u8data=buf->buffer[buf->head];
		buf->head++;
		return 0;
	}
	else return 1;
}

/*
void USARTInit() {
	
	// Set baud rate
	UBRR0H = (uint8_t)(MYUBRR>>8);
	UBRR0L = (uint8_t)MYUBRR;

	// Set frame format to 8 data bits, no parity, 1 stop bit
	UCSR0C |= (1<<UCSZ01) | (1<<UCSZ00);
	UCSR0B |= (1<<RXEN0) | (1<<RXCIE0);
	
}

void enableReading(void) {
	//disable transmission and UDR0 empty interrupt
	UCSR0B &= ~((1<<TXEN0)|(1<<UDRIE0));
	//enable reception and RC complete interrupt
	UCSR0B |= (1<<RXEN0)|(1<<RXCIE0);
}

void enableWriting(void) {
	//disable reception and RX Complete interrupt
	UCSR0B &= ~(
		//(1<<RXEN0)|
		(1<<RXCIE0)
	);
	//enable transmission and UDR0 empty interrupt
	UCSR0B |= (1<<TXEN0)|(1<<UDRIE0);
}

void sendCommand(char* command) {
	//Buffer zurcksetzen, schreiben und dann Transmit-Interrupts enablen
	BufferInit(&buf);
	while(*command != 0x00) {
		BufferWrite(&buf, (*command));
		command++;
	}
	
	BufferWrite(&buf, '\r');
	BufferWrite(&buf, '\n');
	enableWriting();
}

*/
void handleSetRGB(char* command) {
	char* params = strsep(&command, "/");
	char *rgbToken;
	
	int red = 0, green = 0, blue = 0;
	
	while((rgbToken = strsep(&params, "&")) != NULL) {
		
		char color = rgbToken[0];
		int value = atoi(rgbToken+2);
		if (color == 'r') {
			red = value;
			} else if (color == 'g') {
			green = value;
			} else if (color == 'b') {
			blue = value;
		}
		
	}
	
	setRGB(red, green, blue);
}

/*

//RX Complete interrupt service routine
ISR(USART_RX_vect) {
	uint8_t u8temp;
	u8temp=UDR0;
	
	//gar nicht groartig auf die Antwort reagieren, solange die Initialisierung noch gar nicht fertig ist. Etwaige Requests werden dann halt ignoriert
	if (initializationComplete == 0) {
		return;
	}

	//Irgendein OK oder ERROR
	if (BufferWrite(&buf, u8temp) == 1 || strstr(buf.buffer, "OK") != NULL ) {
		blinkR(5);
	//Get-Request
	} else if (strstr(buf.buffer, "CONNECT") != NULL && strstr(buf.buffer, "Accept-Language") != NULL) {
		blinkG(5);
		char* uriRow = strstr(buf.buffer, "GET /rest");
		char channel; 
		
		if (uriRow == NULL ) {
			
		} else {
			blinkG(5);
			char* substr = strstr(buf.buffer, "+IPD,");
			channel = substr[5];
			char* token;
			
			//GET-part ignorieren
			token = strsep(&uriRow, " ");
			//command auslesen
			token = strsep(&uriRow, " ");
			if (token != NULL) {
				char* command = strdup(token+6);
				
				token = strsep(&command, "/");
				
				//weiche f. verschiedene Commands
				if (strcmp(token, "setRGB") == 0) {
					handleSetRGB(command);
				}
				
			}
			
		}
		
		if (channel != '\0') {
			
			char command[14];
			sprintf(command, "%s%c%s", "AT+CIPSEND=", channel, ",2");
			sendCommand(*command);
			_delay_ms(100);
			sendCommand("OK");
			_delay_ms(100);
			sprintf(command, "%s%c", "AT+CIPCLOSE=", channel);
			sendCommand(*command);
		}
		
	}
}

//UDR0 Empty - interrupt service routine
ISR(USART_UDRE_vect) {
	if (BufferRead(&buf, &UDR0) == 1) {
		//start over
		//reset buffer
		BufferInit(&buf);
		enableReading();
	}
}

*/

void setupPwm() {
	//Pins auf Output setzen
	DDRD = 0 | (1 << DDD6) | (1 << DDD5) | (1 << DDD3); //OC0A sitzt auf Pin 12 (PD6), OCOB auf Pin 11 (PD5) und OC2B auf Pin 5 (PD3) -- DDR = Data Direction Register, 0 = Input, 1 = Output
	
	//Timer/Counter 0 Control Register (Register A und B) - 8 Bit each
	//Hier setzt man eigentlich nur die Settings, das hat nichts mit dem rauf/runterzaehlen zu tun. Daher reichen 8 Bit auch vllig aus
	
	/* TCCR0A = TC0 Control Register A */
	//Fr FAST PWM: COM0A1 = 1 und COM0A0 = 1: Set OC0A on compare match, clear OC0A at BOTTOM, (inverting mode)
	//D.h. je hoeher der Wert im TCCR, desto krzer leuchtet die LED
	TCCR0A = 0 | (1 << COM0A1) | (1 << COM0A0);
	
	//Selbiges gilt fuer COM0B1 und COM0B0
	TCCR0A |= (1 << COM0B1) | (1 << COM0B0);
	
	//WGM01 und WGM11 = 1 -> Fast PWM
	TCCR0A |= (1 << WGM01) | (1 << WGM00);
	
	/* TCCR0B TC0 Control Register B */
	//Nicht das Register fuer zweiten Counter, da das schon in TCCR0A mitkonfiguriert ist (COB0B1 und COM0B0).
	
	//FOC0A und FOC0B haben nur bei non-pwm Modes eine Funktion, mssen aber zwecks Kompatibilitt auf 0 gesetzt werden 
	TCCR0B = 0 | (0 << FOC0A) | (0 << FOC0B); 
	
	//WGM02 auf 0, somit ist TOP = 0xff, und der compare match triggert beim Wert von OCR0A bzw. OCR0B. Wenn WGM02 = 1, dann ist top = OCR0A und fr den compare match wird OCR0B herangezogen
	TCCR0B |= (0 << WGM02);
	
	/*
	*	CS12		CS11		CS10		Prescaler
	*	0			0			0			No clock source / timer stopped			
	*	0						1			No prescaling
	*	0			1			0			Prescaler 8
	*	0			1			1			Prescaler 64
		
		mehr dazu im Datasheet auf Seite 173
	*/
	
	TCCR0B |= (0 << CS02) | (0 << CS01) | (1 << CS00);

	/************************************************************************/

	/* Ab hier Timer/Counter 2                                              */

	/************************************************************************/
	
	/* Timer/Counter 2 Control Register A */
	//wenn Fast PWM: inverted mode
	TCCR2A = 0 | (1 << COM2B1) | (1 << COM2B0);
	
	//PWM auf drittem Channel (OC2A) bleibt erstmal deaktiviert, da nicht bentigt
	
	//Fast PWM
	TCCR2A |= (1 << WGM21) | (1 << WGM20);
	
	//Wieder die ungenutzten Flags auf 0 setzen 
	TCCR2B = 0 | (0 < FOC2A) | (0 < FOC2B);
	
	//WGM22 ebenfalls auf 0, damit Register f. compare match Verwendung findet
	TCCR2B |= (0 < WGM22); 
	
	//und ebenfalls kein Prescaler
	TCCR2B |= (0 < CS22) | (0 < CS21) | (1 < CS20);
}

void hsi2rgb(float H, float S, float I, int* rgb) {
	int r, g, b;
	H = fmod(H,360); // cycle H around to 0-360 degrees
	H = 3.14159*H/(float)180; // Convert to radians.
	S = S>0?(S<1?S:1):0; // clamp S and I to interval [0,1]
	I = I>0?(I<1?I:1):0;
	
	// Math! Thanks in part to Kyle Miller.
	if(H < 2.09439) {
		r = 255*I/3*(1+S*cos(H)/cos(1.047196667-H));
		g = 255*I/3*(1+S*(1-cos(H)/cos(1.047196667-H)));
		b = 255*I/3*(1-S);
	} else if(H < 4.188787) {
		H = H - 2.09439;
		g = 255*I/3*(1+S*cos(H)/cos(1.047196667-H));
		b = 255*I/3*(1+S*(1-cos(H)/cos(1.047196667-H)));
		r = 255*I/3*(1-S);
	} else {
		H = H - 4.188787;
		b = 255*I/3*(1+S*cos(H)/cos(1.047196667-H));
		r = 255*I/3*(1+S*(1-cos(H)/cos(1.047196667-H)));
		g = 255*I/3*(1-S);
	}
	rgb[0]=r;
	rgb[1]=g;
	rgb[2]=b;
}

/*
 OC1A, OC1B und OC2 sind die Ausgabepins der PWM                      
 OCR1A etc sind die zugehrigen Compare Registers 
 Dabei wird der Counter inkrementiert und bei Match bit OCRn wird OCn gesetzt.                   
*/

int8_t setupWifiController() {
	int8_t returnValue; 
	char *match; 
	
	//Reset mal, damit das sauber hochstartet
	printf("AT+RST\r\n");
	returnValue = uexpect(stdin, "Ai-Thinker Technology Co.,Ltd.\r\n", NULL); 
	
	_delay_ms(10000);
	//Echo off
	printf("ATE0\r\n");
	returnValue = uexpect(stdin, responseOk, NULL);
	if (returnValue != 0) {
		return returnValue;
	}
	
	//current wifi mode - muss 1 (Station) sein
	printf("AT+CWMODE?\r\n"); 
	returnValue = uexpect(stdin, responseOk, NULL); 
	if (returnValue != 0) {
		return returnValue;
	}
	
	//Den Mode auslesen
	match = strstr(uexpect_before, "+CWMODE:"); 
	if (match == NULL) {
		return -1; 
	}
	
	//wenn nicht schon 1, dann auf 1 setzen 
	if (match[8] != '1') {
		printf("AT+CWMODE=1\r\n");
		returnValue = uexpect(stdin, responseOk, NULL);
		if (returnValue != 0) {
			return returnValue; 
		}
	}
	
	printf("AT+CIPMUX=1\r\n"); 
	returnValue = uexpect(stdin, responseOk, NULL);
	if (returnValue != 0) {
		return returnValue;
	}	
}

int8_t connectToWifi(char *ssid, char* passwd) {
	int8_t returnValue; 
	char *match;
	
	printf("AT+CWJAP?\r\n");
	returnValue = uexpect(stdin, responseOk, NULL);
	if (returnValue != 0) {
		return returnValue;
	}
	
	match = strstr(uexpect_before, "No AP");
	if (match != NULL) {
		printf("AT+CWJAP=\"%s\",\"%s\"\r\n", ssid, passwd);

		returnValue = uexpect(stdin, "\r\nOK\r\n", "\r\nFAIL\r\n", NULL);
		if (returnValue != 0)  {
			return returnValue;
		}
	}
	return 0;
}

int main(void) {
	int8_t returnValue;
	setupPwm();
	 
	uart_init(0);
	stdin = &uart_input; 
	stdout = &uart_output;
	 
	returnValue = setupWifiController();
	if (returnValue != 0) {
		blinkG(100);
		return returnValue;
	}
	 
	returnValue = connectToWifi("UPC1374179", "MWKTJTCV");
	if (returnValue != 0) {
		blinkR(100);
		return returnValue;
	}
	
	returnValue = connectToWifi("UPC1374179", "MWKTJTCV");
	if (returnValue != 0) {
		blinkR(100);
		return returnValue;
	}
	
	//Webserver auf Port 80
	printf("AT+CIPSERVER=1,80\r\n");
	returnValue = uexpect(stdin, responseOk, NULL);
	if (returnValue != 0) {
		return returnValue;
	}
	
	 //Speicher f. RGB-Werte reservieren
	 malloc(3*sizeof(int));
	 float h = 0;
	 float s = 1;
	 float i = 1;
	  
	 
	 while(1) {
		
		hsi2rgb(h, s, i, rgb);
		COUNTER_RED_LED = rgb[0];
		COUNTER_GREEN_LED = rgb[1];
		COUNTER_BLUE_LED = rgb[2];
		h += 0.1;
		_delay_ms(10);
		

	 }
}


